"""Configuration parameters for pytest."""
# pylint: disable=unused-import
import sys
from os.path import abspath
from os.path import dirname as d
from os.path import join

from pytest_mock import mocker
